import sys
import unittest

if __name__ == "__main__":
    tests = unittest.defaultTestLoader.discover("test", pattern="test_*.py")
    runner = unittest.runner.TextTestRunner()
    result = runner.run(tests)
    sys.exit(not result.wasSuccessful())



import contextlib
import gc
import os
import sys
import mprofile
import subprocess
import unittest
from unittest.mock import patch

try:
    import _testcapi
except ImportError:
    _testcapi = None


EMPTY_STRING_SIZE = sys.getsizeof(b"")
INVALID_NFRAME = (-1, 2 ** 30)


def get_frames(nframe, lineno_delta):
    frames = []
    frame = sys._getframe(1)
    for index in range(nframe):
        code = frame.f_code
        lineno = frame.f_lineno + lineno_delta
        frames.append((code.co_name, code.co_filename, code.co_firstlineno, lineno))
        lineno_delta = 0
        frame = frame.f_back
        if frame is None:
            break
    return tuple(frames)


def allocate_bytes(size):
    nframe = mprofile.get_traceback_limit()
    bytes_len = size - EMPTY_STRING_SIZE
    frames = get_frames(nframe, 1)
    data = b"x" * bytes_len
    return data, mprofile.Traceback(frames)


def create_snapshots():
    traceback_limit = 2

    # _mprofile._get_traces() returns a list of (size,
    # traceback_frames) tuples. traceback_frames is a tuple of
    # (name, filename, start_line, line_number) tuples.
    raw_traces = [
        (10, (("test", "a.py", 1, 2), ("test", "b.py", 1, 4))),
        (10, (("test", "a.py", 1, 2), ("test", "b.py", 1, 4))),
        (10, (("test", "a.py", 1, 2), ("test", "b.py", 1, 4))),
        (2, (("test", "a.py", 1, 5), ("test", "b.py", 1, 4))),
        (66, (("test", "b.py", 1, 1),)),
        (7, (("test", "<unknown>", 1, 0),)),
    ]
    snapshot = mprofile.Snapshot(raw_traces, traceback_limit)

    raw_traces2 = [
        (10, (("test", "a.py", 1, 2), ("test", "b.py", 1, 4))),
        (10, (("test", "a.py", 1, 2), ("test", "b.py", 1, 4))),
        (10, (("test", "a.py", 1, 2), ("test", "b.py", 1, 4))),
        (2, (("test", "a.py", 1, 5), ("test", "b.py", 1, 4))),
        (5000, (("test", "a.py", 1, 5), ("test", "b.py", 1, 4))),
        (400, (("test", "c.py", 1, 578),)),
    ]
    snapshot2 = mprofile.Snapshot(raw_traces2, traceback_limit)

    return (snapshot, snapshot2)


def frame(name, filename, firstlineno, lineno):
    return mprofile.Frame((name, filename, firstlineno, lineno))


def traceback(*frames):
    frames = tuple(("test", f[0], 1, f[1]) for f in frames)
    return mprofile.Traceback(frames)


def traceback_lineno(filename, lineno):
    return traceback((filename, lineno))


def traceback_filename(filename):
    frames = (("", filename, 0, 0),)
    return mprofile.Traceback(frames)


class TestMProfileEnabled(unittest.TestCase):
    def setUp(self):
        if mprofile.is_tracing():
            self.skipTest("mprofile must be stopped before the test")

        mprofile.start(1)

    def tearDown(self):
        mprofile.stop()

    def test_get_mprofile_memory(self):
        data = [allocate_bytes(123) for count in range(1000)]
        size = mprofile.get_tracemalloc_memory()
        self.assertGreaterEqual(size, 0)

        mprofile.clear_traces()
        size2 = mprofile.get_tracemalloc_memory()
        self.assertGreaterEqual(size2, 0)
        self.assertLessEqual(size2, size)

    def test_get_object_traceback(self):
        mprofile.clear_traces()
        obj_size = 12345
        obj, obj_traceback = allocate_bytes(obj_size)
        traceback = mprofile.get_object_traceback(obj)
        self.assertEqual(traceback, obj_traceback)

    @unittest.skip("This test fails because there's no way to hook _Py_NewReference")
    def test_new_reference(self):
        mprofile.clear_traces()
        gc.collect()

        # Create a list and "destroy it": put it in the PyListObject free list
        obj = []
        obj = None

        # Create a list which should reuse the previously created empty list
        obj = []

        nframe = mprofile.get_traceback_limit()
        frames = get_frames(nframe, -3)
        obj_traceback = mprofile.Traceback(frames)

        print("Getting reference")
        import sys

        sys.stdout.flush()
        traceback = mprofile.get_object_traceback(obj)
        self.assertIsNotNone(traceback)
        self.assertEqual(traceback, obj_traceback)

    def test_set_traceback_limit(self):
        obj_size = 10

        mprofile.stop()
        self.assertRaises(ValueError, mprofile.start, -1)

        mprofile.stop()
        mprofile.start(10)
        obj2, obj2_traceback = allocate_bytes(obj_size)
        traceback = mprofile.get_object_traceback(obj2)
        self.assertEqual(len(traceback), 10)
        self.assertEqual(traceback, obj2_traceback)

        mprofile.stop()
        mprofile.start(1)
        obj, obj_traceback = allocate_bytes(obj_size)
        traceback = mprofile.get_object_traceback(obj)
        self.assertEqual(len(traceback), 1)
        self.assertEqual(traceback, obj_traceback)

    def find_trace(self, traces, traceback):
        for trace in traces:
            if trace[1] == traceback._frames:
                return trace

        self.fail("trace not found")

    def test_get_traces(self):
        mprofile.clear_traces()
        obj_size = 12345
        obj, obj_traceback = allocate_bytes(obj_size)

        traces = mprofile._get_traces()
        trace = self.find_trace(traces, obj_traceback)

        self.assertIsInstance(trace, tuple)
        size, traceback = trace
        self.assertEqual(size, obj_size)
        self.assertEqual(traceback, obj_traceback._frames)

        mprofile.stop()
        self.assertEqual(mprofile._get_traces(), [])

    def test_get_traces_intern_traceback(self):
        # dummy wrappers to get more useful and identical frames in the traceback
        def allocate_bytes2(size):
            return allocate_bytes(size)

        def allocate_bytes3(size):
            return allocate_bytes2(size)

        def allocate_bytes4(size):
            return allocate_bytes3(size)

        # Ensure that two identical tracebacks are not duplicated
        mprofile.stop()
        mprofile.start(4)
        obj_size = 123
        obj1, obj1_traceback = allocate_bytes4(obj_size)
        obj2, obj2_traceback = allocate_bytes4(obj_size)

        traces = mprofile._get_traces()

        obj1_traceback._frames = tuple(reversed(obj1_traceback._frames))
        obj2_traceback._frames = tuple(reversed(obj2_traceback._frames))

        trace1 = self.find_trace(traces, obj1_traceback)
        trace2 = self.find_trace(traces, obj2_traceback)
        size1, traceback1 = trace1
        size2, traceback2 = trace2
        self.assertIs(traceback2, traceback1)

    def test_get_traced_memory(self):
        # Python allocates some internals objects, so the test must tolerate
        # a small difference between the expected size and the real usage
        max_error = 2048

        # allocate one object
        obj_size = 1024 * 1024
        mprofile.clear_traces()
        obj, obj_traceback = allocate_bytes(obj_size)
        size, peak_size = mprofile.get_traced_memory()
        self.assertGreaterEqual(size, obj_size)
        self.assertGreaterEqual(peak_size, size)

        self.assertLessEqual(size - obj_size, max_error)
        self.assertLessEqual(peak_size - size, max_error)

        # destroy the object
        obj = None
        size2, peak_size2 = mprofile.get_traced_memory()
        self.assertLess(size2, size)
        self.assertGreaterEqual(size - size2, obj_size - max_error)
        self.assertGreaterEqual(peak_size2, peak_size)

        # clear_traces() must reset traced memory counters
        mprofile.clear_traces()
        self.assertEqual(mprofile.get_traced_memory(), (0, 0))

        # allocate another object
        obj, obj_traceback = allocate_bytes(obj_size)
        size, peak_size = mprofile.get_traced_memory()
        self.assertGreaterEqual(size, obj_size)

        # stop() also resets traced memory counters
        mprofile.stop()
        self.assertEqual(mprofile.get_traced_memory(), (0, 0))

    def test_clear_traces(self):
        obj, obj_traceback = allocate_bytes(12345)
        traceback = mprofile.get_object_traceback(obj)
        self.assertIsNotNone(traceback)

        mprofile.clear_traces()
        traceback2 = mprofile.get_object_traceback(obj)
        self.assertIsNone(traceback2)

    def test_is_tracing(self):
        mprofile.stop()
        self.assertFalse(mprofile.is_tracing())

        mprofile.start()
        self.assertTrue(mprofile.is_tracing())

    def test_snapshot(self):
        obj, source = allocate_bytes(123)

        # take a snapshot
        snapshot = mprofile.take_snapshot()

        # mprofile must be tracing memory allocations to take a snapshot
        mprofile.stop()
        with self.assertRaises(RuntimeError) as cm:
            mprofile.take_snapshot()
        self.assertEqual(
            str(cm.exception),
            "the mprofile module must be tracing memory "
            "allocations to take a snapshot",
        )

    def fork_child(self):
        if not mprofile.is_tracing():
            return 2

        obj_size = 12345
        obj, obj_traceback = allocate_bytes(obj_size)
        traceback = mprofile.get_object_traceback(obj)
        if traceback is None:
            return 3

        # everything is fine
        return 0

    @unittest.skipUnless(hasattr(os, "fork"), "need os.fork()")
    def test_fork(self):
        # check that mprofile is still working after fork
        pid = os.fork()
        if not pid:
            # child
            exitcode = 1
            try:
                exitcode = self.fork_child()
            finally:
                os._exit(exitcode)
        else:
            pid2, status = os.waitpid(pid, 0)
            self.assertTrue(os.WIFEXITED(status))
            exitcode = os.WEXITSTATUS(status)
            self.assertEqual(exitcode, 0)


class TestSnapshot(unittest.TestCase):
    maxDiff = 4000

    def test_create_snapshot(self):
        raw_traces = [(5, (("test", "a.py", 1, 2),))]

        with contextlib.ExitStack() as stack:
            stack.enter_context(patch.object(mprofile, "is_tracing", return_value=True))
            stack.enter_context(
                patch.object(mprofile, "get_traceback_limit", return_value=5)
            )
            stack.enter_context(
                patch.object(mprofile, "get_sample_rate", return_value=1)
            )
            stack.enter_context(
                patch.object(mprofile, "_get_traces", return_value=raw_traces)
            )

            snapshot = mprofile.take_snapshot()
            self.assertEqual(snapshot.traceback_limit, 5)
            self.assertEqual(len(snapshot.traces), 1)
            self.assertEqual(snapshot.sample_rate, 1)
            trace = snapshot.traces[0]
            self.assertEqual(trace.size, 5)
            self.assertEqual(len(trace.traceback), 1)
            self.assertEqual(trace.traceback[0].filename, "a.py")
            self.assertEqual(trace.traceback[0].lineno, 2)

    def test_filter_traces(self):
        snapshot, snapshot2 = create_snapshots()
        filter1 = mprofile.Filter(False, "b.py")
        filter2 = mprofile.Filter(True, "a.py", 2)
        filter3 = mprofile.Filter(True, "a.py", 5)

        original_traces = list(snapshot.traces._traces)

        # exclude b.py
        snapshot3 = snapshot.filter_traces((filter1,))
        self.assertEqual(
            snapshot3.traces._traces,
            [
                (10, (("test", "a.py", 1, 2), ("test", "b.py", 1, 4))),
                (10, (("test", "a.py", 1, 2), ("test", "b.py", 1, 4))),
                (10, (("test", "a.py", 1, 2), ("test", "b.py", 1, 4))),
                (2, (("test", "a.py", 1, 5), ("test", "b.py", 1, 4))),
                (7, (("test", "<unknown>", 1, 0),)),
            ],
        )

        # filter_traces() must not touch the original snapshot
        self.assertEqual(snapshot.traces._traces, original_traces)

        # only include two lines of a.py
        snapshot4 = snapshot3.filter_traces((filter2, filter3))
        self.assertEqual(
            snapshot4.traces._traces,
            [
                (10, (("test", "a.py", 1, 2), ("test", "b.py", 1, 4))),
                (10, (("test", "a.py", 1, 2), ("test", "b.py", 1, 4))),
                (10, (("test", "a.py", 1, 2), ("test", "b.py", 1, 4))),
                (2, (("test", "a.py", 1, 5), ("test", "b.py", 1, 4))),
            ],
        )

        # No filter: just duplicate the snapshot
        snapshot5 = snapshot.filter_traces(())
        self.assertIsNot(snapshot5, snapshot)
        self.assertIsNot(snapshot5.traces, snapshot.traces)
        self.assertEqual(snapshot5.traces, snapshot.traces)

        self.assertRaises(TypeError, snapshot.filter_traces, filter1)

    def test_snapshot_group_by_line(self):
        snapshot, snapshot2 = create_snapshots()
        tb_0 = traceback_lineno("<unknown>", 0)
        tb_a_2 = traceback_lineno("a.py", 2)
        tb_a_5 = traceback_lineno("a.py", 5)
        tb_b_1 = traceback_lineno("b.py", 1)
        tb_c_578 = traceback_lineno("c.py", 578)

        # stats per file and line
        stats1 = snapshot.statistics("lineno")
        self.assertEqual(
            stats1,
            [
                mprofile.Statistic(tb_b_1, 66, 1),
                mprofile.Statistic(tb_a_2, 30, 3),
                mprofile.Statistic(tb_0, 7, 1),
                mprofile.Statistic(tb_a_5, 2, 1),
            ],
        )

        # stats per file and line (2)
        stats2 = snapshot2.statistics("lineno")
        self.assertEqual(
            stats2,
            [
                mprofile.Statistic(tb_a_5, 5002, 2),
                mprofile.Statistic(tb_c_578, 400, 1),
                mprofile.Statistic(tb_a_2, 30, 3),
            ],
        )

        # stats diff per file and line
        statistics = snapshot2.compare_to(snapshot, "lineno")
        self.assertEqual(
            statistics,
            [
                mprofile.StatisticDiff(tb_a_5, 5002, 5000, 2, 1),
                mprofile.StatisticDiff(tb_c_578, 400, 400, 1, 1),
                mprofile.StatisticDiff(tb_b_1, 0, -66, 0, -1),
                mprofile.StatisticDiff(tb_0, 0, -7, 0, -1),
                mprofile.StatisticDiff(tb_a_2, 30, 0, 3, 0),
            ],
        )

    def test_snapshot_group_by_file(self):
        snapshot, snapshot2 = create_snapshots()
        tb_0 = traceback_filename("<unknown>")
        tb_a = traceback_filename("a.py")
        tb_b = traceback_filename("b.py")
        tb_c = traceback_filename("c.py")

        # stats per file
        stats1 = snapshot.statistics("filename")
        self.assertEqual(
            stats1,
            [
                mprofile.Statistic(tb_b, 66, 1),
                mprofile.Statistic(tb_a, 32, 4),
                mprofile.Statistic(tb_0, 7, 1),
            ],
        )

        # stats per file (2)
        stats2 = snapshot2.statistics("filename")
        self.assertEqual(
            stats2,
            [mprofile.Statistic(tb_a, 5032, 5), mprofile.Statistic(tb_c, 400, 1)],
        )

        # stats diff per file
        diff = snapshot2.compare_to(snapshot, "filename")
        self.assertEqual(
            diff,
            [
                mprofile.StatisticDiff(tb_a, 5032, 5000, 5, 1),
                mprofile.StatisticDiff(tb_c, 400, 400, 1, 1),
                mprofile.StatisticDiff(tb_b, 0, -66, 0, -1),
                mprofile.StatisticDiff(tb_0, 0, -7, 0, -1),
            ],
        )

    def test_snapshot_group_by_traceback(self):
        snapshot, snapshot2 = create_snapshots()

        # stats per file
        tb1 = traceback(("a.py", 2), ("b.py", 4))
        tb2 = traceback(("a.py", 5), ("b.py", 4))
        tb3 = traceback(("b.py", 1))
        tb4 = traceback(("<unknown>", 0))
        stats1 = snapshot.statistics("traceback")
        self.assertEqual(
            stats1,
            [
                mprofile.Statistic(tb3, 66, 1),
                mprofile.Statistic(tb1, 30, 3),
                mprofile.Statistic(tb4, 7, 1),
                mprofile.Statistic(tb2, 2, 1),
            ],
        )

        # stats per file (2)
        tb5 = traceback(("c.py", 578))
        stats2 = snapshot2.statistics("traceback")
        self.assertEqual(
            stats2,
            [
                mprofile.Statistic(tb2, 5002, 2),
                mprofile.Statistic(tb5, 400, 1),
                mprofile.Statistic(tb1, 30, 3),
            ],
        )

        # stats diff per file
        diff = snapshot2.compare_to(snapshot, "traceback")
        self.assertEqual(
            diff,
            [
                mprofile.StatisticDiff(tb2, 5002, 5000, 2, 1),
                mprofile.StatisticDiff(tb5, 400, 400, 1, 1),
                mprofile.StatisticDiff(tb3, 0, -66, 0, -1),
                mprofile.StatisticDiff(tb4, 0, -7, 0, -1),
                mprofile.StatisticDiff(tb1, 30, 0, 3, 0),
            ],
        )

        self.assertRaises(ValueError, snapshot.statistics, "traceback", cumulative=True)

    def test_snapshot_group_by_cumulative(self):
        snapshot, snapshot2 = create_snapshots()
        tb_0 = traceback_filename("<unknown>")
        tb_a = traceback_filename("a.py")
        tb_b = traceback_filename("b.py")
        tb_0_0 = traceback_lineno("<unknown>", 0)
        tb_a_2 = traceback_lineno("a.py", 2)
        tb_a_5 = traceback_lineno("a.py", 5)
        tb_b_1 = traceback_lineno("b.py", 1)
        tb_b_4 = traceback_lineno("b.py", 4)

        # per file
        stats = snapshot.statistics("filename", True)
        self.assertEqual(
            stats,
            [
                mprofile.Statistic(tb_b, 98, 5),
                mprofile.Statistic(tb_a, 32, 4),
                mprofile.Statistic(tb_0, 7, 1),
            ],
        )

        # per line
        stats = snapshot.statistics("lineno", True)
        self.assertEqual(
            stats,
            [
                mprofile.Statistic(tb_b_1, 66, 1),
                mprofile.Statistic(tb_b_4, 32, 4),
                mprofile.Statistic(tb_a_2, 30, 3),
                mprofile.Statistic(tb_0_0, 7, 1),
                mprofile.Statistic(tb_a_5, 2, 1),
            ],
        )

    def test_trace_format(self):
        snapshot, snapshot2 = create_snapshots()
        trace = snapshot.traces[0]
        self.assertEqual(str(trace), "b.py:4: 10 B")
        traceback = trace.traceback
        self.assertEqual(str(traceback), "b.py:4")
        frame = traceback[0]
        self.assertEqual(str(frame), "b.py:4")

    def test_statistic_format(self):
        snapshot, snapshot2 = create_snapshots()
        stats = snapshot.statistics("lineno")
        stat = stats[0]
        self.assertEqual(str(stat), "b.py:1: size=66 B, count=1, average=66 B")

    def test_statistic_diff_format(self):
        snapshot, snapshot2 = create_snapshots()
        stats = snapshot2.compare_to(snapshot, "lineno")
        stat = stats[0]
        self.assertEqual(
            str(stat), "a.py:5: size=5002 B (+5000 B), count=2 (+1), average=2501 B"
        )

    def test_slices(self):
        snapshot, snapshot2 = create_snapshots()
        self.assertEqual(snapshot.traces[:2], (snapshot.traces[0], snapshot.traces[1]))

        traceback = snapshot.traces[0].traceback
        self.assertEqual(traceback[:2], (traceback[0], traceback[1]))

    def test_format_traceback(self):
        snapshot, snapshot2 = create_snapshots()

        def getline(filename, lineno):
            return "  <%s, %s>" % (filename, lineno)

        with unittest.mock.patch("mprofile.linecache.getline", side_effect=getline):
            tb = snapshot.traces[0].traceback
            self.assertEqual(
                tb.format(),
                [
                    '  File "b.py", line 4',
                    "    <b.py, 4>",
                    '  File "a.py", line 2',
                    "    <a.py, 2>",
                ],
            )

            self.assertEqual(
                tb.format(limit=1), ['  File "a.py", line 2', "    <a.py, 2>"]
            )

            self.assertEqual(
                tb.format(limit=-1), ['  File "b.py", line 4', "    <b.py, 4>"]
            )

            self.assertEqual(
                tb.format(most_recent_first=True),
                [
                    '  File "a.py", line 2',
                    "    <a.py, 2>",
                    '  File "b.py", line 4',
                    "    <b.py, 4>",
                ],
            )

            self.assertEqual(
                tb.format(limit=1, most_recent_first=True),
                ['  File "a.py", line 2', "    <a.py, 2>"],
            )

            self.assertEqual(
                tb.format(limit=-1, most_recent_first=True),
                ['  File "b.py", line 4', "    <b.py, 4>"],
            )


class TestFilters(unittest.TestCase):
    maxDiff = 2048

    def test_filter_attributes(self):
        # test default values
        f = mprofile.Filter(True, "abc")
        self.assertEqual(f.inclusive, True)
        self.assertEqual(f.filename_pattern, "abc")
        self.assertIsNone(f.lineno)
        self.assertEqual(f.all_frames, False)

        # test custom values
        f = mprofile.Filter(False, "test.py", 123, True)
        self.assertEqual(f.inclusive, False)
        self.assertEqual(f.filename_pattern, "test.py")
        self.assertEqual(f.lineno, 123)
        self.assertEqual(f.all_frames, True)

        # parameters passed by keyword
        f = mprofile.Filter(
            inclusive=False, filename_pattern="test.py", lineno=123, all_frames=True
        )
        self.assertEqual(f.inclusive, False)
        self.assertEqual(f.filename_pattern, "test.py")
        self.assertEqual(f.lineno, 123)
        self.assertEqual(f.all_frames, True)

        # read-only attribute
        self.assertRaises(AttributeError, setattr, f, "filename_pattern", "abc")

    def test_filter_match(self):
        # filter without line number
        f = mprofile.Filter(True, "abc")
        self.assertTrue(f._match_frame("abc", 0))
        self.assertTrue(f._match_frame("abc", 5))
        self.assertTrue(f._match_frame("abc", 10))
        self.assertFalse(f._match_frame("12356", 0))
        self.assertFalse(f._match_frame("12356", 5))
        self.assertFalse(f._match_frame("12356", 10))

        f = mprofile.Filter(False, "abc")
        self.assertFalse(f._match_frame("abc", 0))
        self.assertFalse(f._match_frame("abc", 5))
        self.assertFalse(f._match_frame("abc", 10))
        self.assertTrue(f._match_frame("12356", 0))
        self.assertTrue(f._match_frame("12356", 5))
        self.assertTrue(f._match_frame("12356", 10))

        # filter with line number > 0
        f = mprofile.Filter(True, "abc", 5)
        self.assertFalse(f._match_frame("abc", 0))
        self.assertTrue(f._match_frame("abc", 5))
        self.assertFalse(f._match_frame("abc", 10))
        self.assertFalse(f._match_frame("12356", 0))
        self.assertFalse(f._match_frame("12356", 5))
        self.assertFalse(f._match_frame("12356", 10))

        f = mprofile.Filter(False, "abc", 5)
        self.assertTrue(f._match_frame("abc", 0))
        self.assertFalse(f._match_frame("abc", 5))
        self.assertTrue(f._match_frame("abc", 10))
        self.assertTrue(f._match_frame("12356", 0))
        self.assertTrue(f._match_frame("12356", 5))
        self.assertTrue(f._match_frame("12356", 10))

        # filter with line number 0
        f = mprofile.Filter(True, "abc", 0)
        self.assertTrue(f._match_frame("abc", 0))
        self.assertFalse(f._match_frame("abc", 5))
        self.assertFalse(f._match_frame("abc", 10))
        self.assertFalse(f._match_frame("12356", 0))
        self.assertFalse(f._match_frame("12356", 5))
        self.assertFalse(f._match_frame("12356", 10))

        f = mprofile.Filter(False, "abc", 0)
        self.assertFalse(f._match_frame("abc", 0))
        self.assertTrue(f._match_frame("abc", 5))
        self.assertTrue(f._match_frame("abc", 10))
        self.assertTrue(f._match_frame("12356", 0))
        self.assertTrue(f._match_frame("12356", 5))
        self.assertTrue(f._match_frame("12356", 10))

    def test_filter_match_filename(self):
        def fnmatch(inclusive, filename, pattern):
            f = mprofile.Filter(inclusive, pattern)
            return f._match_frame(filename, 0)

        self.assertTrue(fnmatch(True, "abc", "abc"))
        self.assertFalse(fnmatch(True, "12356", "abc"))
        self.assertFalse(fnmatch(True, "<unknown>", "abc"))

        self.assertFalse(fnmatch(False, "abc", "abc"))
        self.assertTrue(fnmatch(False, "12356", "abc"))
        self.assertTrue(fnmatch(False, "<unknown>", "abc"))

    def test_filter_match_filename_joker(self):
        def fnmatch(filename, pattern):
            filter = mprofile.Filter(True, pattern)
            return filter._match_frame(filename, 0)

        # empty string
        self.assertFalse(fnmatch("abc", ""))
        self.assertFalse(fnmatch("", "abc"))
        self.assertTrue(fnmatch("", ""))
        self.assertTrue(fnmatch("", "*"))

        # no *
        self.assertTrue(fnmatch("abc", "abc"))
        self.assertFalse(fnmatch("abc", "abcd"))
        self.assertFalse(fnmatch("abc", "def"))

        # a*
        self.assertTrue(fnmatch("abc", "a*"))
        self.assertTrue(fnmatch("abc", "abc*"))
        self.assertFalse(fnmatch("abc", "b*"))
        self.assertFalse(fnmatch("abc", "abcd*"))

        # a*b
        self.assertTrue(fnmatch("abc", "a*c"))
        self.assertTrue(fnmatch("abcdcx", "a*cx"))
        self.assertFalse(fnmatch("abb", "a*c"))
        self.assertFalse(fnmatch("abcdce", "a*cx"))

        # a*b*c
        self.assertTrue(fnmatch("abcde", "a*c*e"))
        self.assertTrue(fnmatch("abcbdefeg", "a*bd*eg"))
        self.assertFalse(fnmatch("abcdd", "a*c*e"))
        self.assertFalse(fnmatch("abcbdefef", "a*bd*eg"))

        # replace .pyc suffix with .py
        self.assertTrue(fnmatch("a.pyc", "a.py"))
        self.assertTrue(fnmatch("a.py", "a.pyc"))

        if os.name == "nt":
            # case insensitive
            self.assertTrue(fnmatch("aBC", "ABc"))
            self.assertTrue(fnmatch("aBcDe", "Ab*dE"))

            self.assertTrue(fnmatch("a.pyc", "a.PY"))
            self.assertTrue(fnmatch("a.py", "a.PYC"))
        else:
            # case sensitive
            self.assertFalse(fnmatch("aBC", "ABc"))
            self.assertFalse(fnmatch("aBcDe", "Ab*dE"))

            self.assertFalse(fnmatch("a.pyc", "a.PY"))
            self.assertFalse(fnmatch("a.py", "a.PYC"))

        if os.name == "nt":
            # normalize alternate separator "/" to the standard separator "\"
            self.assertTrue(fnmatch(r"a/b", r"a\b"))
            self.assertTrue(fnmatch(r"a\b", r"a/b"))
            self.assertTrue(fnmatch(r"a/b\c", r"a\b/c"))
            self.assertTrue(fnmatch(r"a/b/c", r"a\b\c"))
        else:
            # there is no alternate separator
            self.assertFalse(fnmatch(r"a/b", r"a\b"))
            self.assertFalse(fnmatch(r"a\b", r"a/b"))
            self.assertFalse(fnmatch(r"a/b\c", r"a\b/c"))
            self.assertFalse(fnmatch(r"a/b/c", r"a\b\c"))

        self.assertTrue(fnmatch("a.pyo", "a.py"))

    def test_filter_match_trace(self):
        t1 = (("test", "a.py", 1, 2), ("test", "b.py", 1, 3))
        t2 = (("test", "b.py", 1, 4), ("test", "b.py", 1, 5))
        t3 = (("test", "c.py", 1, 5), ("test", "<unknown>", 0, 0))
        unknown = (("test", "<unknown>", 0, 0),)

        f = mprofile.Filter(True, "b.py", all_frames=True)
        self.assertTrue(f._match_traceback(t1))
        self.assertTrue(f._match_traceback(t2))
        self.assertFalse(f._match_traceback(t3))
        self.assertFalse(f._match_traceback(unknown))

        f = mprofile.Filter(True, "b.py", all_frames=False)
        self.assertFalse(f._match_traceback(t1))
        self.assertTrue(f._match_traceback(t2))
        self.assertFalse(f._match_traceback(t3))
        self.assertFalse(f._match_traceback(unknown))

        f = mprofile.Filter(False, "b.py", all_frames=True)
        self.assertFalse(f._match_traceback(t1))
        self.assertFalse(f._match_traceback(t2))
        self.assertTrue(f._match_traceback(t3))
        self.assertTrue(f._match_traceback(unknown))

        f = mprofile.Filter(False, "b.py", all_frames=False)
        self.assertTrue(f._match_traceback(t1))
        self.assertFalse(f._match_traceback(t2))
        self.assertTrue(f._match_traceback(t3))
        self.assertTrue(f._match_traceback(unknown))

        f = mprofile.Filter(False, "<unknown>", all_frames=False)
        self.assertTrue(f._match_traceback(t1))
        self.assertTrue(f._match_traceback(t2))
        self.assertTrue(f._match_traceback(t3))
        self.assertFalse(f._match_traceback(unknown))

        f = mprofile.Filter(True, "<unknown>", all_frames=True)
        self.assertFalse(f._match_traceback(t1))
        self.assertFalse(f._match_traceback(t2))
        self.assertTrue(f._match_traceback(t3))
        self.assertTrue(f._match_traceback(unknown))

        f = mprofile.Filter(False, "<unknown>", all_frames=True)
        self.assertTrue(f._match_traceback(t1))
        self.assertTrue(f._match_traceback(t2))
        self.assertFalse(f._match_traceback(t3))
        self.assertFalse(f._match_traceback(unknown))


class TestCommandLine(unittest.TestCase):
    def test_env_var_disabled_by_default(self):
        # not tracing by default
        code = "import mprofile; print(mprofile.is_tracing())"
        stdout = subprocess.check_output([sys.executable, "-c", code])
        stdout = stdout.rstrip()
        self.assertEqual(stdout, b"False")

    def test_env_var_enabled_at_startup(self):
        # tracing at startup
        code = "import mprofile; print(mprofile.is_tracing())"
        stdout = subprocess.check_output(
            [sys.executable, "-c", code], env={"MPROFILERATE": "1"}
        )
        stdout = stdout.rstrip()
        self.assertEqual(stdout, b"True")

    def test_env_limit(self):
        # start and set the number of frames
        code = "import mprofile; print(mprofile.get_traceback_limit())"
        stdout = subprocess.check_output(
            [sys.executable, "-c", code],
            env={"MPROFILERATE": "1", "MPROFILEFRAMES": "10"},
        )
        stdout = stdout.rstrip()
        self.assertEqual(stdout, b"10")

    def check_env_var_invalid(self, nframe):
        code = "import mprofile; print(mprofile.is_tracing())"
        with self.assertRaises(subprocess.CalledProcessError) as cm:
            subprocess.check_output(
                [sys.executable, "-c", code],
                env={"MPROFILERATE": "1", "MPROFILEFRAMES": str(nframe)},
                stderr=subprocess.STDOUT,
            )

        stdout = cm.exception.output
        if b"ValueError: the number of frames must be in range" in stdout:
            return
        if b"MPROFILEFRAMES: invalid number of frames" in stdout:
            return
        self.fail("unexpected output: {}".format(stdout))

    def test_env_var_invalid(self):
        for nframe in INVALID_NFRAME:
            with self.subTest(nframe=nframe):
                self.check_env_var_invalid(nframe)

    @unittest.skipIf(_testcapi is None, "need _testcapi")
    def test_pymem_alloc0(self):
        # Issue #21639: Check that PyMem_Malloc(0) with mprofile enabled
        # does not crash.
        code = "import _testcapi; _testcapi.test_pymem_alloc0(); 1"
        stdout = subprocess.check_output(
            [sys.executable, "-c", code], env={"MPROFILERATE": "1"}
        )



# This file is largely adapted from tracemalloc.
from collections import Sequence, Iterable
import fnmatch
from functools import total_ordering
import linecache
import math
import os.path

# Import types and functions implemented in C
try:
    from mprofile._profiler import *
    from mprofile._profiler import _get_object_traceback, _get_traces

    _ext_available = True
except ImportError as e:
    _ext_available = False

    def is_tracing():
        return False


# setup.py reads the version information from here to set package version
__version__ = "0.0.12"


def _assert_ext_available():
    if not _ext_available:
        raise RuntimeError(
            "mprofile: C extension is not available, you must use Python >= 3.4 "
            + "or a patched interpreter"
        )


def _format_size(size, sign):
    for unit in ("B", "KiB", "MiB", "GiB", "TiB"):
        if abs(size) < 100 and unit != "B":
            # 3 digits (xx.x UNIT)
            if sign:
                return "%+.1f %s" % (size, unit)
            else:
                return "%.1f %s" % (size, unit)
        if abs(size) < 10 * 1024 or unit == "TiB":
            # 4 or 5 digits (xxxx UNIT)
            if sign:
                return "%+.0f %s" % (size, unit)
            else:
                return "%.0f %s" % (size, unit)
        size /= 1024


class Statistic(object):
    """
    Statistic difference on memory allocations between two Snapshot instance.
    """

    __slots__ = ("traceback", "size", "count")

    def __init__(self, traceback, size, count):
        self.traceback = traceback
        self.size = size
        self.count = count

    def __hash__(self):
        return hash((self.traceback, self.size, self.count))

    def __eq__(self, other):
        return (
            self.traceback == other.traceback
            and self.size == other.size
            and self.count == other.count
        )

    def __str__(self):
        text = "%s: size=%s, count=%i" % (
            self.traceback,
            _format_size(self.size, False),
            self.count,
        )
        if self.count:
            average = self.size / self.count
            text += ", average=%s" % _format_size(average, False)
        return text

    def __repr__(self):
        return "<Statistic traceback=%r size=%i count=%i>" % (
            self.traceback,
            self.size,
            self.count,
        )

    def _sort_key(self):
        return (self.size, self.count, self.traceback)


class StatisticDiff(object):
    """
    Statistic difference on memory allocations between an old and a new
    Snapshot instance.
    """

    __slots__ = ("traceback", "size", "size_diff", "count", "count_diff")

    def __init__(self, traceback, size, size_diff, count, count_diff):
        self.traceback = traceback
        self.size = size
        self.size_diff = size_diff
        self.count = count
        self.count_diff = count_diff

    def __hash__(self):
        return hash(
            (self.traceback, self.size, self.size_diff, self.count, self.count_diff)
        )

    def __eq__(self, other):
        return (
            self.traceback == other.traceback
            and self.size == other.size
            and self.size_diff == other.size_diff
            and self.count == other.count
            and self.count_diff == other.count_diff
        )

    def __str__(self):
        text = "%s: size=%s (%s), count=%i (%+i)" % (
            self.traceback,
            _format_size(self.size, False),
            _format_size(self.size_diff, True),
            self.count,
            self.count_diff,
        )
        if self.count:
            average = self.size / self.count
            text += ", average=%s" % _format_size(average, False)
        return text

    def __repr__(self):
        return "<StatisticDiff traceback=%r size=%i (%+i) count=%i (%+i)>" % (
            self.traceback,
            self.size,
            self.size_diff,
            self.count,
            self.count_diff,
        )

    def _sort_key(self):
        return (
            abs(self.size_diff),
            self.size,
            abs(self.count_diff),
            self.count,
            self.traceback,
        )


def _compare_grouped_stats(old_group, new_group):
    statistics = []
    for traceback, stat in new_group.items():
        previous = old_group.pop(traceback, None)
        if previous is not None:
            stat = StatisticDiff(
                traceback,
                stat.size,
                stat.size - previous.size,
                stat.count,
                stat.count - previous.count,
            )
        else:
            stat = StatisticDiff(
                traceback, stat.size, stat.size, stat.count, stat.count
            )
        statistics.append(stat)

    for traceback, stat in old_group.items():
        stat = StatisticDiff(traceback, 0, -stat.size, 0, -stat.count)
        statistics.append(stat)
    return statistics


@total_ordering
class Frame(object):
    """
    Frame of a traceback.
    """

    __slots__ = ("_frame",)

    def __init__(self, frame):
        # frame is a tuple: (name: str, filename: str, firstlineno: int, lineno: int)
        self._frame = frame

    @property
    def name(self):
        return self._frame[0]

    @property
    def filename(self):
        return self._frame[1]

    @property
    def firstlineno(self):
        return self._frame[2]

    @property
    def lineno(self):
        return self._frame[3]

    def __eq__(self, other):
        return self._frame == other._frame

    def __lt__(self, other):
        return self._frame < other._frame

    def __hash__(self):
        return hash(self._frame)

    def __str__(self):
        return "%s:%s" % (self.filename, self.lineno)

    def __repr__(self):
        return "<Frame filename=%r name=%r (%r) lineno=%r>" % (
            self.filename,
            self.name,
            self.firstlineno,
            self.lineno,
        )


@total_ordering
class Traceback(Sequence):
    """
    Sequence of Frame instances sorted from the most recent frame
    to the oldest frame.
    """

    __slots__ = ("_frames",)

    def __init__(self, frames):
        Sequence.__init__(self)
        # frames is a tuple of frame tuples: see Frame constructor for the
        # format of a frame tuple
        self._frames = tuple(reversed(frames))

    def __len__(self):
        return len(self._frames)

    def __getitem__(self, index):
        if isinstance(index, slice):
            return tuple(Frame(trace) for trace in self._frames[index])
        else:
            return Frame(self._frames[index])

    def __contains__(self, frame):
        return frame._frame in self._frames

    def __hash__(self):
        return hash(self._frames)

    def __eq__(self, other):
        return self._frames == other._frames

    def __lt__(self, other):
        return self._frames < other._frames

    def __str__(self):
        return str(self[0])

    def __repr__(self):
        return "<Traceback %r>" % (tuple(self),)

    def format(self, limit=None, most_recent_first=False):
        lines = []
        if limit is not None:
            if limit > 0:
                frame_slice = self[-limit:]
            else:
                frame_slice = self[:limit]
        else:
            frame_slice = self

        if most_recent_first:
            frame_slice = reversed(frame_slice)
        for frame in frame_slice:
            lines.append('  File "%s", line %s' % (frame.filename, frame.lineno))
            line = linecache.getline(frame.filename, frame.lineno).strip()
            if line:
                lines.append("    %s" % line)
        return lines


def get_object_traceback(obj):
    """
    Get the traceback where the Python object *obj* was allocated.
    Return a Traceback instance.

    Return None if the mprofile module is not tracing memory allocations or
    did not trace the allocation of the object.
    """
    _assert_ext_available()
    frames = _get_object_traceback(obj)
    if frames:
        return Traceback(frames)
    else:
        return None


class Trace(object):
    """
    Trace of a memory block.
    """

    __slots__ = ("_trace",)

    def __init__(self, trace):
        # trace is a tuple: (size, traceback), see Traceback constructor
        # for the format of the traceback tuple
        self._trace = trace

    @property
    def traceback(self):
        return Traceback(self._trace[1])

    @property
    def size(self):
        return self._trace[0]

    def __eq__(self, other):
        return self._trace == other._trace

    def __hash__(self):
        return hash(self._trace)

    def __str__(self):
        return "%s: %s" % (self.traceback, _format_size(self.size, False))

    def __repr__(self):
        return "<Trace size=%s, traceback=%r>" % (
            _format_size(self.size, False),
            self.traceback,
        )


class _Traces(Sequence):
    def __init__(self, traces):
        Sequence.__init__(self)
        # traces is a tuple of trace tuples: see Trace constructor
        self._traces = traces

    def __len__(self):
        return len(self._traces)

    def __getitem__(self, index):
        if isinstance(index, slice):
            return tuple(Trace(trace) for trace in self._traces[index])
        else:
            return Trace(self._traces[index])

    def __contains__(self, trace):
        return trace._trace in self._traces

    def __eq__(self, other):
        return self._traces == other._traces

    def __repr__(self):
        return "<Traces len=%s>" % len(self)


def _normalize_filename(filename):
    filename = os.path.normcase(filename)
    if filename.endswith((".pyc", ".pyo")):
        filename = filename[:-1]
    return filename


class Filter(object):
    def __init__(self, inclusive, filename_pattern, lineno=None, all_frames=False):
        self.inclusive = inclusive
        self._filename_pattern = _normalize_filename(filename_pattern)
        self.lineno = lineno
        self.all_frames = all_frames

    @property
    def filename_pattern(self):
        return self._filename_pattern

    def __match_frame(self, filename, lineno):
        filename = _normalize_filename(filename)
        if not fnmatch.fnmatch(filename, self._filename_pattern):
            return False
        if self.lineno is None:
            return True
        else:
            return lineno == self.lineno

    def _match_frame(self, filename, lineno):
        return self.__match_frame(filename, lineno) ^ (not self.inclusive)

    def _match_traceback(self, traceback):
        if self.all_frames:
            if any(
                self.__match_frame(filename, lineno)
                for _, filename, _, lineno in traceback
            ):
                return self.inclusive
            else:
                return not self.inclusive
        else:
            _, filename, _, lineno = traceback[0]
            return self._match_frame(filename, lineno)


class Snapshot(object):
    """
    Snapshot of traces of memory blocks allocated by Python.
    """

    def __init__(self, traces, traceback_limit, sample_rate=0):
        # traces is a tuple of trace tuples: see _Traces constructor for
        # the exact format
        self.traces = _Traces(traces)
        self.traceback_limit = traceback_limit
        self.sample_rate = sample_rate

    def _filter_trace(self, include_filters, exclude_filters, trace):
        traceback = trace[1]
        if include_filters:
            if not any(
                trace_filter._match_traceback(traceback)
                for trace_filter in include_filters
            ):
                return False
        if exclude_filters:
            if any(
                not trace_filter._match_traceback(traceback)
                for trace_filter in exclude_filters
            ):
                return False
        return True

    def filter_traces(self, filters):
        """
        Create a new Snapshot instance with a filtered traces sequence, filters
        is a list of Filter instances.  If filters is an empty list, return a
        new Snapshot instance with a copy of the traces.
        """
        if not isinstance(filters, Iterable):
            raise TypeError(
                "filters must be a list of filters, not %s" % type(filters).__name__
            )
        if filters:
            include_filters = []
            exclude_filters = []
            for trace_filter in filters:
                if trace_filter.inclusive:
                    include_filters.append(trace_filter)
                else:
                    exclude_filters.append(trace_filter)
            new_traces = [
                trace
                for trace in self.traces._traces
                if self._filter_trace(include_filters, exclude_filters, trace)
            ]
        else:
            new_traces = self.traces._traces[:]
        return Snapshot(new_traces, self.traceback_limit, self.sample_rate)

    def _group_by(self, key_type, cumulative):
        if key_type not in ("traceback", "filename", "lineno"):
            raise ValueError("unknown key_type: %r" % (key_type,))
        if cumulative and key_type not in ("lineno", "filename"):
            raise ValueError(
                "cumulative mode cannot by used " "with key type %r" % key_type
            )

        stats = {}
        tracebacks = {}
        if not cumulative:
            for trace in self.traces._traces:
                size, trace_traceback = trace
                try:
                    traceback = tracebacks[trace_traceback]
                except KeyError:
                    if key_type == "traceback":
                        frames = trace_traceback
                    elif key_type == "lineno":
                        frames = trace_traceback[:1]
                    else:  # key_type == 'filename':
                        # Synthetic frame with just the filename.
                        frames = (("", trace_traceback[0][1], 0, 0),)
                    traceback = Traceback(frames)
                    tracebacks[trace_traceback] = traceback
                try:
                    stat = stats[traceback]
                    stat.size += size
                    stat.count += 1
                except KeyError:
                    stats[traceback] = Statistic(traceback, size, 1)
        else:
            # cumulative statistics
            for trace in self.traces._traces:
                size, trace_traceback = trace
                for frame in trace_traceback:
                    try:
                        traceback = tracebacks[frame]
                    except KeyError:
                        if key_type == "lineno":
                            frames = (frame,)
                        else:  # key_type == 'filename':
                            frames = (("", frame[1], 0, 0),)
                        traceback = Traceback(frames)
                        tracebacks[frame] = traceback
                    try:
                        stat = stats[traceback]
                        stat.size += size
                        stat.count += 1
                    except KeyError:
                        stats[traceback] = Statistic(traceback, size, 1)
        return self._scale_heap_samples(stats)

    def _scale_heap_samples(self, stats):
        for tb, stat in stats.items():
            stats[tb] = self._scale_heap_sample(stat)
        return stats

    def _scale_heap_sample(self, stat):
        if stat.count == 0 or stat.size == 0:
            return stat
        if self.sample_rate <= 1:
            return stat
        avg_size = float(stat.size) / stat.count
        scale = 1.0 / (1.0 - math.exp(-avg_size / self.sample_rate))
        stat.size = int(scale * stat.size)
        stat.count = int(scale * stat.count)
        return stat

    def statistics(self, key_type, cumulative=False):
        """
        Group statistics by key_type. Return a sorted list of Statistic
        instances.
        """
        grouped = self._group_by(key_type, cumulative)
        statistics = list(grouped.values())
        statistics.sort(reverse=True, key=Statistic._sort_key)
        return statistics

    def compare_to(self, old_snapshot, key_type, cumulative=False):
        """
        Compute the differences with an old snapshot old_snapshot. Get
        statistics as a sorted list of StatisticDiff instances, grouped by
        group_by.
        """
        new_group = self._group_by(key_type, cumulative)
        old_group = old_snapshot._group_by(key_type, cumulative)
        statistics = _compare_grouped_stats(old_group, new_group)
        statistics.sort(reverse=True, key=StatisticDiff._sort_key)
        return statistics


def take_snapshot():
    """
    Take a snapshot of traces of memory blocks allocated by Python.
    """
    _assert_ext_available()
    if not is_tracing():
        raise RuntimeError(
            "the mprofile module must be tracing memory "
            "allocations to take a snapshot"
        )
    traces = _get_traces()
    traceback_limit = get_traceback_limit()
    sample_rate = get_sample_rate()
    return Snapshot(traces, traceback_limit, sample_rate)
